@@ -3,38 +3,39 @@
 """
 *The recipe to generate a master dark frame*
 
-:Author:
-    David Young & Marco Landoni
+Author
+: David Young & Marco Landoni
 
-:Date Created:
-    January 27, 2020
+Date Created
+: January 27, 2020
 """
 ################# GLOBAL IMPORTS ####################
 from soxspipe.commonutils import keyword_lookup
 
-from ._base_recipe_ import _base_recipe_
+from .base_recipe import base_recipe
 
 from fundamentals import tools
 from builtins import object
 from datetime import datetime
 from soxspipe.commonutils.toolkit import generic_quality_checks
 import sys
-import math
 import os
 os.environ['TERM'] = 'vt100'
 
 
-class soxs_mdark(_base_recipe_):
+class soxs_mdark(base_recipe):
     """
-    *The soxs_mdark recipe*
+    *The `soxs_mdark` recipe generates a master-dark frame used to remove flux attributed to the dark current from other frames.*
 
     **Key Arguments**
 
-        - ``log`` -- logger
-        - ``settings`` -- the settings dictionary
-        - ``inputFrames`` -- input fits frames. Can be a directory, a set-of-files (SOF) file or a list of fits frame paths.
-        - ``verbose`` -- verbose. True or False. Default *False*
-        - ``overwrite`` -- overwrite the prodcut file if it already exists. Default *False*
+    - ``log`` -- logger
+    - ``settings`` -- the settings dictionary
+    - ``inputFrames`` -- input fits frames. Can be a directory, a set-of-files (SOF) file or a list of fits frame paths.
+    - ``verbose`` -- verbose. True or False. Default *False*
+    - ``overwrite`` -- overwrite the product file if it already exists. Default *False*
+    - ``command`` -- the command called to run the recipe
+
 
     **Usage**
 
@@ -43,16 +44,10 @@ class soxs_mdark(_base_recipe_):
     mdarkFrame = soxs_mdark(
         log=log,
         settings=settings,
-        inputFrames=fileList
-    )..produce_product()
-    ```
-
-    ---
-
-    ```eval_rst
-    .. todo::
-
-        - add a tutorial about ``soxs_mdark`` to documentation
+        inputFrames=fileList,
+        verbose=False,
+        overwrite=False
+    ).produce_product()
     ```
     """
 
@@ -62,17 +57,17 @@ class soxs_mdark(_base_recipe_):
             settings=False,
             inputFrames=[],
             verbose=False,
-            overwrite=False
+            overwrite=False,
+            command=False
 
     ):
-        # INHERIT INITIALISATION FROM  _base_recipe_
-        super(soxs_mdark, self).__init__(log=log, settings=settings, inputFrames=inputFrames, overwrite=overwrite, recipeName="soxs-mdark")
+        # INHERIT INITIALISATION FROM  base_recipe
+        super(soxs_mdark, self).__init__(log=log, settings=settings, inputFrames=inputFrames, overwrite=overwrite, recipeName="soxs-mdark", command=command)
         self.log = log
-        log.debug("instansiating a new 'soxs_mdark' object")
+        log.debug("instantiating a new 'soxs_mdark' object")
         self.settings = settings
         self.inputFrames = inputFrames
         self.verbose = verbose
-        self.recipeSettings = settings[self.recipeName]
         # xt-self-arg-tmpx
 
         # INITIAL ACTIONS
@@ -89,16 +84,17 @@ class soxs_mdark(_base_recipe_):
 
         # VERIFY THE FRAMES ARE THE ONES EXPECTED BY SOXS_MDARK - NO MORE, NO LESS.
         # PRINT SUMMARY OF FILES.
-        print("# VERIFYING INPUT FRAMES")
+        self.log.print("# VERIFYING INPUT FRAMES")
         self.verify_input_frames()
+        sys.stdout.flush()
         sys.stdout.write("\x1b[1A\x1b[2K")
-        print("# VERIFYING INPUT FRAMES - ALL GOOD")
+        self.log.print("# VERIFYING INPUT FRAMES - ALL GOOD")
 
         # SORT IMAGE COLLECTION
         self.inputFrames.sort(['MJD-OBS'])
         if self.verbose:
-            print("# RAW INPUT FRAMES - SUMMARY")
-            print(self.inputFrames.summary, "\n")
+            self.log.print("# RAW INPUT FRAMES - SUMMARY")
+            self.log.print(self.inputFrames.summary, "\n")
 
         # PREPARE THE FRAMES - CONVERT TO ELECTRONS, ADD UNCERTAINTY AND MASK
         # EXTENSIONS
@@ -111,7 +107,7 @@ class soxs_mdark(_base_recipe_):
             self):
         """*verify input frame match those required by the soxs_mdark recipe*
 
-        If the fits files conform to required input for the recipe everything will pass silently, otherwise an exception shall be raised.
+        If the fits files conform to the required input for the recipe, everything will pass silently; otherwise, an exception will be raised.
         """
         self.log.debug('starting the ``verify_input_frames`` method')
 
@@ -120,27 +116,33 @@ class soxs_mdark(_base_recipe_):
         # BASIC VERIFICATION COMMON TO ALL RECIPES
         imageTypes, imageTech, imageCat = self._verify_input_frames_basics()
 
-        # MIXED INPUT IMAGE TYPES ARE BAD
-        if len(imageTypes) > 1:
-            imageTypes = " and ".join(imageTypes)
-            print(self.inputFrames.summary)
-            raise TypeError(
-                "Input frames are a mix of %(imageTypes)s" % locals())
-        # NON-BIAS INPUT IMAGE TYPES ARE BAD
-        elif imageTypes[0] != 'DARK':
-            print(self.inputFrames.summary)
-            raise TypeError(
-                "Input frames not DARK frames" % locals())
+        error = False
 
-        exptimes = self.inputFrames.values(
-            keyword=kw("EXPTIME"), unique=True)
         # MIXED INPUT IMAGE TYPES ARE BAD
-        if len(exptimes) > 1:
-            exptimes = [str(e) for e in exptimes]
-            exptimes = " and ".join(exptimes)
-            print(self.inputFrames.summary)
-            raise TypeError(
-                "Input frames have differing exposure-times %(exptimes)s" % locals())
+        if not error:
+            if len(imageTypes) > 1:
+                imageTypes = " and ".join(imageTypes)
+                error = "Input frames are a mix of %(imageTypes)s" % locals()
+            # NON-BIAS INPUT IMAGE TYPES ARE BAD
+            elif imageTypes[0] != 'DARK':
+                error = "Input frames not DARK frames" % locals()
+
+        if not error:
+            exptimes = self.inputFrames.values(
+                keyword=kw("EXPTIME"), unique=True)
+            # MIXED INPUT IMAGE TYPES ARE BAD
+            if len(exptimes) > 1:
+                exptimes = [str(e) for e in exptimes]
+                exptimes = " and ".join(exptimes)
+                error = "Input frames have differing exposure-times %(exptimes)s" % locals()
+
+        if error:
+            sys.stdout.flush()
+            sys.stdout.write("\x1b[1A\x1b[2K")
+            self.log.print("# VERIFYING INPUT FRAMES - **ERROR**\n")
+            self.log.print(self.inputFrames.summary)
+            self.log.print("")
+            raise TypeError(error)
 
         self.imageType = imageTypes[0]
         self.log.debug('completed the ``verify_input_frames`` method')
@@ -151,11 +153,13 @@ class soxs_mdark(_base_recipe_):
         """*generate a master dark frame*
 
         **Return:**
-            - ``productPath`` -- the path to master dark frame
+
+        - ``productPath`` -- the path to master dark frame
         """
         self.log.debug('starting the ``produce_product`` method')
 
         import numpy as np
+        import pandas as pd
 
         arm = self.arm
         kw = self.kw
@@ -164,7 +168,7 @@ class soxs_mdark(_base_recipe_):
         # LIST OF CCDDATA OBJECTS
         ccds = [c for c in self.inputFrames.ccds(ccd_kwargs={"hdu_uncertainty": 'ERRS', "hdu_mask": 'QUAL', "hdu_flags": 'FLAGS', "key_uncertainty_type": 'UTYPE'})]
 
-        meanFluxLevels, rons, noiseFrames = zip(*[self.subtact_mean_flux_level(c) for c in ccds])
+        meanFluxLevels, rons, noiseFrames = zip(*[self.subtract_mean_flux_level(c) for c in ccds])
         masterMeanFluxLevel = np.mean(meanFluxLevels)
         masterMedianFluxLevel = np.median(meanFluxLevels)
         rawRon = np.mean(rons)
@@ -203,7 +207,7 @@ class soxs_mdark(_base_recipe_):
         # WRITE TO DISK
         productPath = self._write(
             frame=combined_dark_mean,
-            filedir=self.intermediateRootPath,
+            filedir=self.workspaceRootPath,
             filename=False,
             overwrite=True
         )
@@ -214,7 +218,7 @@ class soxs_mdark(_base_recipe_):
 
         self.dateObs = combined_dark_mean.header[kw("DATE_OBS")]
 
-        self.products = self.products.append({
+        self.products = pd.concat([self.products, pd.Series({
             "soxspipe_recipe": self.recipeName,
             "product_label": "MDARK",
             "file_name": filename,
@@ -222,8 +226,9 @@ class soxs_mdark(_base_recipe_):
             "obs_date_utc": self.dateObs,
             "reduction_date_utc": utcnow,
             "product_desc": f"{self.arm} Master dark frame",
-            "file_path": productPath
-        }, ignore_index=True)
+            "file_path": productPath,
+            "label": "PROD"
+        }).to_frame().T], ignore_index=True)
 
         self.report_output()
         self.clean_up()
